.line {
    margin-bottom: 1em;
}

.word {
    display: inline-block;
    margin-right: .5em;
    vertical-align: top;
    cursor: pointer;
    color: #202020;
}

.uncorrected {
    color: #ff8000;
}
.miscorrected {
    color: #ff0000;
}
.corrected {
    color: #00a030;
}
.selected {
    background-color: #ddf;
    font-weight: bold;
}

.tip {
    text-align: left;
}

.results-table {
    font-size: small;
}
